#include "usart.h"

USART::USART::USART(void){}
USART::USART::~USART(void){}
//User still needs to handle USART Interrupt activation (within ARM core) if fifo is used
void USART::USART::Init(uint32_t baudrate, USART_TypeDef *USART_InstanceParameter)
{
	Instance = USART_InstanceParameter;
#ifdef USART_USE_FIFO
	FifoInit();
#endif
	//We enable the transmitter and the receiver (the transmitter Interupt gets enabled when we actually need to send something)
	Instance->CR1 = USART_CR1_TE | USART_CR1_RE;
	//We send the MSB first
	Instance->CR2 = 0;
	//Driver enable
	Instance->CR3 = 0;
	//Baudrate of 9600 with an 8MHz Clock
	Instance->BRR = F_CPU/baudrate;
	Instance->CR1 |= USART_CR1_UE;
}

void USART::SpecificInitDefault(void)
{
	//Enable Clocks
	RCC->APB2ENR |= RCC_APB2ENR_USART1EN;
	RCC->AHBENR |= RCC_AHBENR_GPIOAEN;
	//First clear all existing Configuration for the pins we use
	GPIOA->MODER &= ~(GPIO_MODER_MODER9 | GPIO_MODER_MODER10);
	//Then configure them as AF Pins
	GPIOA->MODER |= GPIO_MODER_MODER9_1 | GPIO_MODER_MODER10_1;
	//Clear the AF
	GPIOA->AFR[1] &= ~(0b1111<<(4*(9-8))) |(0b1111<<(4*(10-8)));
	//Write our own configuration
	//They are both AF1, so the value for that is shifted for both pins
	GPIOA->AFR[1] |= (0b0001<<(4*(9-8))) |(0b0001<<(4*(10-8)));

}
#ifdef USART_USE_FIFO
void USART::USART::FifoInit(void)
{
	//Init Ring Buffer
	FifoBuffer.head = 0;
	FifoBuffer.tail = 0;
}

uint32_t USART::USART::FifoSize(void)
{
	return (((FifoBuffer.head - FifoBuffer.tail) - 1) % USART_FifoBufferLength);
}

USART::USART_Status USART::USART::FifoPut(uint8_t data)
{
	if(FifoSize() < 1){
		return USART_ReturnFailed;
	}
	FifoBuffer.buffer[FifoBuffer.tail++] = data;
	if(FifoBuffer.tail >= USART_FifoBufferLength){
		FifoBuffer.tail = 0;
	}
	return USART_ReturnSuccess;
}

USART::USART_Status USART::USART::FifoGet(uint8_t &data)
{
	if(FifoSize() == (USART_FifoBufferLength-1)){
		return USART_ReturnFailed;
	}
	data = FifoBuffer.buffer[FifoBuffer.head++];
	if(FifoBuffer.head >= USART_FifoBufferLength){
		FifoBuffer.head = 0;
	}
	return USART_ReturnSuccess;
}

//Is  meant to be called by the USART Interrupt
void USART::USART::InterruptHandler(void)
{
	if(Instance->ISR & USART_ISR_TXE) {
		uint8_t data;
		//If we still have data to transmit, transmit it.
		//Else stop
		if(FifoGet(data) == USART_ReturnSuccess){
			Instance->TDR = data;
		} else {
			Instance->CR1 &= ~USART_CR1_TXEIE;
		}
	}

}
#endif
USART::USART_Status USART::USART::TransmitByte(uint8_t data)
{
#ifdef USART_USE_FIFO
	if (FifoPut(data) == USART_ReturnFailed){
		return USART_ReturnFailed;
	} else if(!(Instance->CR1 & USART_CR1_TXEIE)){
		//If the TXEIE bit is not set, we currently can't transmit anything, so we enable it
		Instance->CR1 |= USART_CR1_TXEIE;
	}
#else
	while(!(Instance->ISR & USART_ISR_TXE));
	Instance->TDR = data;
#endif
	return USART_ReturnSuccess;
}

uint8_t USART::USART::ReceiveByte(void)
{
	while(!(Instance->ISR & USART_ISR_RXNE));
	return Instance->RDR;
}

bool USART::USART::ReceiveAvailable(void)
{
	return (Instance->ISR & USART_ISR_RXNE);
}

USART::USART_Status USART::USART::PrintString(const char myString[])
{
	uint8_t length;
	for(length = 0; myString[length];length++);
#ifdef USART_USE_FIFO
	//for length checking
	if (FifoSize() < length ) {
		return USART_ReturnFailed;
	}
#endif
	//If length available
	for(uint8_t i = 0; i < length; i++){
		TransmitByte(myString[i]);
	}
	return USART_ReturnSuccess;
}

uint8_t USART::USART::ReadString(char myString[], uint8_t maxLength)
{
	char response;
	uint8_t i;
	i = 0;
	while (i < (maxLength - 1)) {                   /* prevent over-runs */
		response = ReceiveByte();
		TransmitByte(response);                                    /* echo */
		if (response == '\r') {                     /* enter marks the end */
			break;
		}
		else {
			myString[i] = response;                       /* add in a letter */
			i++;
		}
	}
	myString[i] = 0;                          /* terminal NULL character */
	return i;
}

void USART::USART::PrintByte(uint8_t byte)
{
	/* Converts a byte to a string of decimal text, sends it */
	TransmitByte('0' + (byte / 100));                        /* Hundreds */
	TransmitByte('0' + ((byte / 10) % 10));                      /* Tens */
	TransmitByte('0' + (byte % 10));                             /* Ones */
}

USART::USART_Status USART::USART::PrintWord(uint16_t word)
{
#ifdef USART_USE_FIFO
	if(FifoSize() < 5)
		return USART_ReturnFailed;
#endif
	TransmitByte('0' + (word / 10000));                 /* Ten-thousands */
	TransmitByte('0' + ((word / 1000) % 10));               /* Thousands */
	TransmitByte('0' + ((word / 100) % 10));                 /* Hundreds */
	TransmitByte('0' + ((word / 10) % 10));                      /* Tens */
	TransmitByte('0' + (word % 10));                             /* Ones */
	return USART_ReturnSuccess;
}

void USART::USART::PrintBinaryByte(uint8_t byte)
{
	/* Prints out a byte as a series of 1's and 0's */
	uint8_t bit;
	for (bit = 7; bit < 255; bit--) {
		if (byte && bit)
			TransmitByte('1');
		else
			TransmitByte('0');
	}
}

char USART::USART::NibbleToHexCharacter(uint8_t nibble)
{
	/* Converts 4 bits into hexadecimal */
	if (nibble < 10) {
		return ('0' + nibble);
	}
	else {
		return ('A' + nibble - 10);
	}
}

uint8_t USART::USART::HexCharacterToNibble(char hex)
{
	if(hex >= '0' && '9' >= hex) {
		return hex - '0';
	} else if (hex >= 'A' && 'F' >= hex) {
		return (hex - 'A') + 10;
	} else {
		return 0;
	}

}
void USART::USART::PrintHexByte(uint8_t byte)
{
	/* Prints a byte as its hexadecimal equivalent */
	uint8_t nibble;
	nibble = (byte & 0b11110000) >> 4;
	TransmitByte(NibbleToHexCharacter(nibble));
	nibble = byte & 0b00001111;
	TransmitByte(NibbleToHexCharacter(nibble));
}

uint8_t USART::USART::ReadHexByte(void)
{
	uint8_t byte;
	uint8_t input;
	input = ReceiveByte();
	TransmitByte(input);
	byte = (HexCharacterToNibble(input)<<4);
	input = ReceiveByte();
	TransmitByte(input);
	byte |= HexCharacterToNibble(input);
	return byte;
}

uint8_t USART::USART::GetNumber(void)
{
	// Gets a numerical 0-255 from the serial port.
	// Converts from string to number.
	char hundreds = '0';
	char tens = '0';
	char ones = '0';
	char thisChar = '0';
	do {                                                   /* shift over */
		hundreds = tens;
		tens = ones;
		ones = thisChar;
		thisChar = ReceiveByte();                   /* get a new character */
		TransmitByte(thisChar);                                    /* echo */
	} while (thisChar != '\r');                     /* until type return */
	return (100 * (hundreds - '0') + 10 * (tens - '0') + ones - '0');
}
